package cn.jisol.game.controller;

import cn.hutool.core.util.IdUtil;
import cn.hutool.core.util.RandomUtil;
import cn.jisol.game.data.Cache;
import cn.jisol.game.dto.LoginDTO;
import cn.jisol.game.entity.User;
import cn.jisol.game.service.UserService;
import cn.jisol.game.service.impl.UserServiceImpl;
import cn.jisol.ngame.util.NewsContext;
import com.baomidou.mybatisplus.core.conditions.query.LambdaQueryWrapper;
import com.baomidou.mybatisplus.core.toolkit.Wrappers;
import io.swagger.annotations.Api;
import io.swagger.annotations.ApiImplicitParams;
import io.swagger.annotations.ApiOperation;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.web.bind.annotation.*;

import java.util.Objects;

/**
 * 玩家 API
 */
@Api(value = "JNGameDemo - API", tags = {"PET - API"})
@RestController
@RequestMapping("/user")
@ResponseBody
public class UserController {

    @Autowired
    private UserServiceImpl userService;

    //注册账户接口
    @ApiImplicitParams({})
    @ApiOperation(value = "注册账号")
    @PostMapping("/register")
    public NewsContext<User> register(){
        User user = User.builder().userPass(RandomUtil.randomString(8)).userName("玩家名称").build();
        this.userService.save(user);
        return NewsContext.onSuccess("保存成功",user);
    }

    //登录账户
    @ApiImplicitParams({})
    @ApiOperation(value = "登录账号")
    @PostMapping("/login")
    public NewsContext<LoginDTO> login(@RequestBody User user){
        if(Objects.isNull(user.getUserId()) || Objects.isNull(user.getUserPass())){
            return NewsContext.onFail("登录失败");
        }
        LambdaQueryWrapper<User> query = Wrappers.lambdaQuery();
        query.eq(User::getUserId,user.getUserId());
        query.eq(User::getUserPass,user.getUserPass());
        if(Objects.nonNull(user = userService.getOne(query))){
            user.setUserPass("");
            //生成Token
            String token = IdUtil.simpleUUID();
            Cache.TOKEN.put(token,user);
            return NewsContext.onSuccess("登录成功", LoginDTO.builder().user(user).token(token).build());
        }else{
            return NewsContext.onFail("登录失败");
        }
    }

}
